#lang racket
(require "util.rkt" rackunit)

(define (parse lines)
  (define (ticket->numbers ticket)
    (map string->number (string-split ticket ",")))
  (match-let ([(list ticket-field-rules
                     my-ticket
                     other-tickets) (lines->blocks lines)])
    (values ticket-field-rules
            (map ticket->numbers (rest my-ticket))
            (map ticket->numbers (rest other-tickets)))))

(define-values (ticket-field-rules my-ticket other-tickets)
  (parse (input->lines 16)))

(define (make-ticket-rules ticket-field-rules)
  (define (make-ticket-rule match)
    (λ (x)
      (let* ([clusters (rest match)]
             [name (first clusters)]
             [args (rest clusters)]
             [in-between? (λ (b t) (and (>= x b) (<= x t)))]
             [valid? (λ (b1 t1 b2 t2) (or (in-between? b1 t1)
                                          (in-between? b2 t2)))])
        (if (apply valid? (map string->number args))
            name
            #f))))
  (define pattern #px"([^:]+): (\\d+)-(\\d+) or (\\d+)-(\\d+)")
  (map (compose1 make-ticket-rule ((curry regexp-match) pattern))
       ticket-field-rules))

(define (make-ticket-valid? rules)
  (λ (ticket)
    (if (andmap  (λ (n) (for/first ([r rules])
                         (r n)))
                ticket)
        ticket
        #f)))

(define (make-field-invalid? ticket-field-rules)
  (define rules (make-ticket-rules ticket-field-rules))
  (λ (field)
    (if (for/and ([r rules])
          (not (r field)))
        field
        #f)))

(define (make-ticket-invalid? ticket-field-rules)
  (define field-invalid? (make-field-invalid? ticket-field-rules))
  (λ (ticket)
    (filter-map field-invalid? ticket)))

(define (solve-first lines)
  (define-values (ticket-field-rules _ other-tickets)
    (parse lines))
  (define ticket-invalid? (make-ticket-invalid? ticket-field-rules))
  (define invalid-fields (filter (compose1 not empty?)
                                 (map ticket-invalid?
                                      other-tickets)))
  (apply + (flatten invalid-fields)))

(define (tickets->fields tickets)
  (apply ((curry map) list) tickets))

(define (find-field-names rules other-tickets)
  (define valid-tickets (filter-map (make-ticket-valid? rules) other-tickets))
  (define fields (tickets->fields valid-tickets))
  (define (rule-valid-on-all-fields fields)
    (filter (λ (x) x)
            (for/list ([r rules])
                       (andmap r fields))))

 (define possible-field-names
   (for/list ([f fields]
              [i (in-naturals)])
      (list (rule-valid-on-all-fields f) i)))

  (define (sort-fields fields)
    (sort fields (λ (x y) (< (length (first x))
                             (length (first y))))))
  (define (remove-used-field-name fields name)
    (map (λ (x)
           (list (remove name (first x))
                 (second x)))
         fields))
  ;; (define field-names (for/fold ([acc '()]
  ;;            [sorted (sort-fields possible-field-names)]
  ;;            #:result (map second (sort acc (λ (x y) (< (first x)
  ;;                                                       (first y))))))
  ;;           ([_ (in-list possible-field-names)])
  ;;   (let ([name (let ([names (first (first sorted))])
  ;;                 (if (= (length names) 1)
  ;;                   (first names)
  ;;                   (begin
  ;;                     (displayln sorted)
  ;;                     (error (string-append "More than one possible field name!"
  ;;                                         (~a (first sorted)))) )) )]
  ;;         [key (second (first sorted))])
  ;;     (values (cons (list key name) acc)
  ;;             (sort-fields (remove-used-field-name (rest sorted) name))))))
  (sort-fields possible-field-names ))


(define (solve-second lines)
  (define-values (rule-strings my-ticket other-tickets)
    (parse lines))
  (define ticket-rules (make-ticket-rules rule-strings))
  (define field-names (find-field-names ticket-rules other-tickets))
  field-names)

;; (define-values (input-rules input-my-ticket input-other-tickets)
;;   (parse (input->lines 16)))

;; (define input-ticket-rules (make-ticket-rules input-rules))

;; (define testtttt (for ([i (range 20)]) (let* ([rule-index 0]
;;                         [field-index 0]
;;                         [rule (list-ref input-ticket-rules rule-index)]
;;                         [field (list-ref (tickets->fields input-other-tickets) i)])
;;                    (begin
;;                      (displayln (list-ref input-rules rule-index))
;;                      (map rule field))) ))

(module+ test
  (define test-lines (list "class: 1-3 or 5-7"
                           "row: 6-11 or 33-44"
                           "seat: 13-40 or 45-50"
                           ""
                           "your ticket:"
                           "7,1,14"
                           ""
                           "nearby tickets:"
                           "7,3,47"
                           "40,4,50"
                           "55,2,20"
                           "38,6,12"))
  (define-values (rules my other)
    (parse test-lines))
  (define ticket-invalid? (make-ticket-invalid? rules))
  (define field-invalid? (make-field-invalid? rules))
  (check-equal? (solve-first test-lines) 71)
  (check-equal? (ticket-invalid? (list-ref other 0)) '())
  (check-equal? (ticket-invalid? (list-ref other 1)) '(4))
  (check-equal? (ticket-invalid? (list-ref other 2)) '(55))
  (check-equal? (ticket-invalid? (list-ref other 3)) '(12))
  (define test-lines2 (list "class: 0-1 or 4-19"
                            "row: 0-5 or 8-19"
                            "seat: 0-13 or 16-19"
                            ""
                            "your ticket:"
                            "11,12,13"
                            ""
                            "nearby tickets:"
                            "3,9,18"
                            "15,1,5"
                            "5,14,9"
                            ))
  (define-values (rules2 my2 other2)
    (parse test-lines2))
  (define ticket-rules (make-ticket-rules rules2))
  )
